@@ -1,13 +1,11 @@
 /****************************************************************************/
 // Eclipse SUMO, Simulation of Urban MObility; see https://eclipse.org/sumo
-// Copyright (C) 2001-2017 German Aerospace Center (DLR) and others.
-/****************************************************************************/
-//
-//   This program and the accompanying materials
-//   are made available under the terms of the Eclipse Public License v2.0
-//   which accompanies this distribution, and is available at
-//   http://www.eclipse.org/legal/epl-v20.html
-//
+// Copyright (C) 2001-2018 German Aerospace Center (DLR) and others.
+// This program and the accompanying materials
+// are made available under the terms of the Eclipse Public License v2.0
+// which accompanies this distribution, and is available at
+// http://www.eclipse.org/legal/epl-v20.html
+// SPDX-License-Identifier: EPL-2.0
 /****************************************************************************/
 /// @file    MESegment.cpp
 /// @author  Daniel Krajzewicz
@@ -21,11 +19,7 @@
 // ===========================================================================
 // included modules
 // ===========================================================================
-#ifdef _MSC_VER
-#include <windows_config.h>
-#else
 #include <config.h>
-#endif
 
 #include <algorithm>
 #include <limits>
@@ -53,6 +47,13 @@
 // avoid division by zero when driving very slowly
 #define MESO_MIN_SPEED (0.05)
 
+//#define DEBUG_OPENED
+//#define DEBUG_JAMTHRESHOLD
+//#define DEBUG_COND (getID() == "blocker")
+//#define DEBUG_COND (true)
+#define DEBUG_COND (myEdge.isSelected())
+#define DEBUG_COND2(obj) ((obj != 0 && (obj)->isSelected()))
+
 // ===========================================================================
 // static member defintion
 // ===========================================================================
@@ -83,17 +84,18 @@ MESegment::MESegment(const std::string& id,
     myJunctionControl(junctionControl),
     myTLSPenalty(MSGlobals::gMesoTLSPenalty > 0 &&
                  // only apply to the last segment of a tls-controlled edge
-                 myNextSegment == 0 && (
+                 myNextSegment == nullptr && (
                      parent.getToJunction()->getType() == NODETYPE_TRAFFIC_LIGHT ||
                      parent.getToJunction()->getType() == NODETYPE_TRAFFIC_LIGHT_NOJUNCTION ||
                      parent.getToJunction()->getType() == NODETYPE_TRAFFIC_LIGHT_RIGHT_ON_RED)),
     myMinorPenalty(MSGlobals::gMesoMinorPenalty > 0 &&
                    // only apply to the last segment of an uncontrolled edge that has at least 1 minor link
-                   myNextSegment == 0 &&
+                   myNextSegment == nullptr &&
                    parent.getToJunction()->getType() != NODETYPE_TRAFFIC_LIGHT &&
                    parent.getToJunction()->getType() != NODETYPE_TRAFFIC_LIGHT_NOJUNCTION &&
                    parent.getToJunction()->getType() != NODETYPE_TRAFFIC_LIGHT_RIGHT_ON_RED &&
                    parent.hasMinorLink()),
+    myNumCars(0),
     myEntryBlockTime(SUMOTime_MIN),
     myLastHeadway(TIME2STEPS(-1)),
     myMeanSpeed(speed),
@@ -124,7 +126,7 @@ MESegment::MESegment(const std::string& id,
 MESegment::MESegment(const std::string& id):
     Named(id),
     myEdge(myDummyParent), // arbitrary edge needed to supply the needed reference
-    myNextSegment(0), myLength(0), myIndex(0),
+    myNextSegment(nullptr), myLength(0), myIndex(0),
     myTau_ff(0), myTau_fj(0), myTau_jf(0), myTau_jj(0), myTau_length(1),
     myHeadwayCapacity(0), myCapacity(0), myJunctionControl(false),
     myTLSPenalty(false),
@@ -196,6 +198,12 @@ MESegment::jamThresholdForSpeed(double speed, double jamThresh) const {
     if (speed == 0) {
         return std::numeric_limits<double>::max();  // never jam. Irrelevant at speed 0 anyway
     }
+#ifdef DEBUG_JAMTHRESHOLD
+    if (true || DEBUG_COND) {
+        std::cout << "jamThresholdForSpeed seg=" << getID() << " speed=" << speed << " jamThresh=" << jamThresh << " ffVehs=" << std::ceil(myLength / (-jamThresh * speed * STEPS2TIME(tauWithVehLength(myTau_ff, DEFAULT_VEH_LENGHT_WITH_GAP)))) << " thresh=" << std::ceil(myLength / (-jamThresh * speed * STEPS2TIME(tauWithVehLength(myTau_ff, DEFAULT_VEH_LENGHT_WITH_GAP)))) * DEFAULT_VEH_LENGHT_WITH_GAP
+                  << "\n";
+    }
+#endif
     return std::ceil(myLength / (-jamThresh * speed * STEPS2TIME(tauWithVehLength(myTau_ff, DEFAULT_VEH_LENGHT_WITH_GAP)))) * DEFAULT_VEH_LENGHT_WITH_GAP;
 }
 
@@ -226,22 +234,6 @@ MESegment::removeDetector(MSMoveReminder* data) {
 }
 
 
-void
-MESegment::updateDetectorsOnLeave(MEVehicle* v, SUMOTime currentTime, MESegment* next) {
-    MSMoveReminder::Notification reason;
-    if (next == 0) {
-        reason = MSMoveReminder::NOTIFICATION_ARRIVED;
-    } else if (next == &myVaporizationTarget) {
-        reason = MSMoveReminder::NOTIFICATION_VAPORIZED;
-    } else if (myNextSegment == 0) {
-        reason = MSMoveReminder::NOTIFICATION_JUNCTION;
-    } else {
-        reason = MSMoveReminder::NOTIFICATION_SEGMENT;
-    }
-    v->updateDetectors(currentTime, true, reason);
-}
-
-
 void
 MESegment::prepareDetectorForWriting(MSMoveReminder& data) {
     const SUMOTime currentTime = MSNet::getInstance()->getCurrentTimeStep();
@@ -298,16 +290,6 @@ MESegment::initialise(MEVehicle* veh, SUMOTime time) {
 }
 
 
-int
-MESegment::getCarNumber() const {
-    int total = 0;
-    for (Queues::const_iterator k = myCarQues.begin(); k != myCarQues.end(); ++k) {
-        total += (int)k->size();
-    }
-    return total;
-}
-
-
 double
 MESegment::getMeanSpeed(bool useCached) const {
     const SUMOTime currentTime = MSNet::getInstance()->getCurrentTimeStep();
@@ -345,12 +327,13 @@ MESegment::writeVehicles(OutputDevice& of) const {
 
 
 MEVehicle*
-MESegment::removeCar(MEVehicle* v, SUMOTime leaveTime, MESegment* next) {
+MESegment::removeCar(MEVehicle* v, SUMOTime leaveTime, const MSMoveReminder::Notification reason) {
     myOccupancy = MAX2(0., myOccupancy - v->getVehicleType().getLengthWithGap());
     std::vector<MEVehicle*>& cars = myCarQues[v->getQueIndex()];
     assert(std::find(cars.begin(), cars.end(), v) != cars.end());
     // One could be tempted to do  v->setSegment(next); here but position on lane will be invalid if next == 0
-    updateDetectorsOnLeave(v, leaveTime, next);
+    v->updateDetectors(leaveTime, true, reason);
+    myNumCars--;
     myEdge.lock();
     if (v == cars.back()) {
         cars.pop_back();
@@ -362,7 +345,7 @@ MESegment::removeCar(MEVehicle* v, SUMOTime leaveTime, MESegment* next) {
         cars.erase(std::find(cars.begin(), cars.end(), v));
     }
     myEdge.unlock();
-    return 0;
+    return nullptr;
 }
 
 
@@ -394,8 +377,8 @@ MSLink*
 MESegment::getLink(const MEVehicle* veh, bool penalty) const {
     if (myJunctionControl || penalty) {
         const MSEdge* const nextEdge = veh->succEdge(1);
-        if (nextEdge == 0) {
-            return 0;
+        if (nextEdge == nullptr) {
+            return nullptr;
         }
         // try to find any link leading to our next edge, start with the lane pointed to by the que index
         const MSLane* const bestLane = myEdge.getLanes()[veh->getQueIndex()];
@@ -417,18 +400,40 @@ MESegment::getLink(const MEVehicle* veh, bool penalty) const {
             }
         }
     }
-    return 0;
+    return nullptr;
 }
 
 
 bool
 MESegment::isOpen(const MEVehicle* veh) const {
+#ifdef DEBUG_OPENED
+    if (DEBUG_COND || DEBUG_COND2(veh)) {
+        std::cout << SIMTIME << " opened seg=" << getID() << " veh=" << Named::getIDSecure(veh)
+                  << " tlsPenalty=" << myTLSPenalty;
+        const MSLink* link = getLink(veh);
+        if (link == 0) {
+            std::cout << " link=0";
+        } else {
+            std::cout << " prio=" << link->havePriority()
+                      << " override=" << limitedControlOverride(link)
+                      << " isOpen=" << link->opened(veh->getEventTime(), veh->getSpeed(), veh->estimateLeaveSpeed(link),
+                                                    veh->getVehicleType().getLengthWithGap(), veh->getImpatience(),
+                                                    veh->getVehicleType().getCarFollowModel().getMaxDecel(), veh->getWaitingTime())
+                      << " et=" << veh->getEventTime()
+                      << " v=" << veh->getSpeed()
+                      << " vLeave=" << veh->estimateLeaveSpeed(link)
+                      << " impatience=" << veh->getImpatience()
+                      << " tWait=" << veh->getWaitingTime();
+        }
+        std::cout << "\n";
+    }
+#endif
     if (myTLSPenalty) {
         // XXX should limited control take precedence over tls penalty?
         return true;
     }
     const MSLink* link = getLink(veh);
-    return (link == 0
+    return (link == nullptr
             || link->havePriority()
             || limitedControlOverride(link)
             || link->opened(veh->getEventTime(), veh->getSpeed(), veh->estimateLeaveSpeed(link),
@@ -446,27 +451,30 @@ MESegment::limitedControlOverride(const MSLink* link) const {
     // if the target segment of this link is not saturated junction control is disabled
     const MSEdge& targetEdge = link->getLane()->getEdge();
     const MESegment* target = MSGlobals::gMesoNet->getSegmentForEdge(targetEdge);
-    return target->myOccupancy * 2 < target->myJamThreshold;
+    return (target->myOccupancy * 2 < target->myJamThreshold) && !targetEdge.isRoundabout();
 }
 
 
 void
-MESegment::send(MEVehicle* veh, MESegment* next, SUMOTime time) {
+MESegment::send(MEVehicle* veh, MESegment* next, SUMOTime time, const MSMoveReminder::Notification reason) {
     assert(isInvalid(next) || time >= myBlockTimes[veh->getQueIndex()]);
     MSLink* link = getLink(veh);
-    if (link != 0) {
+    if (link != nullptr) {
         link->removeApproaching(veh);
     }
-    MEVehicle* lc = removeCar(veh, time, next); // new leaderCar
+    MEVehicle* lc = removeCar(veh, time, reason); // new leaderCar
     myBlockTimes[veh->getQueIndex()] = time;
     if (!isInvalid(next)) {
         myLastHeadway = next->getTimeHeadway(this, veh);
         myBlockTimes[veh->getQueIndex()] += myLastHeadway;
     }
-    if (lc != 0) {
+    if (lc != nullptr) {
         lc->setEventTime(MAX2(lc->getEventTime(), myBlockTimes[veh->getQueIndex()]));
         MSGlobals::gMesoNet->addLeaderCar(lc, getLink(lc));
     }
+    if (veh->isStopped()) {
+        veh->processStop();
+    }
 }
 
 bool
@@ -497,7 +505,7 @@ MESegment::receive(MEVehicle* veh, SUMOTime time, bool isDepart, bool afterTelep
         veh->setEventTime(time + TIME2STEPS(myLength / speed)); // for correct arrival speed
         addReminders(veh);
         veh->activateReminders(MSMoveReminder::NOTIFICATION_JUNCTION);
-        updateDetectorsOnLeave(veh, time, 0);
+        veh->updateDetectors(time, true, MSMoveReminder::NOTIFICATION_ARRIVED);
         MSNet::getInstance()->getVehicleControl().scheduleVehicleRemoval(veh);
         return;
     }
@@ -508,7 +516,7 @@ MESegment::receive(MEVehicle* veh, SUMOTime time, bool isDepart, bool afterTelep
     if (myCarQues.size() > 1) {
         const MSEdge* succ = veh->succEdge(1);
         // succ may be invalid if called from initialise() with an invalid route
-        if (succ != 0 && myFollowerMap.count(succ) > 0) {
+        if (succ != nullptr && myFollowerMap.count(succ) > 0) {
             const std::vector<int>& indices = myFollowerMap[succ];
             nextQueIndex = indices[0];
             for (std::vector<int>::const_iterator i = indices.begin() + 1; i != indices.end(); ++i) {
@@ -519,8 +527,11 @@ MESegment::receive(MEVehicle* veh, SUMOTime time, bool isDepart, bool afterTelep
         }
     }
     std::vector<MEVehicle*>& cars = myCarQues[nextQueIndex];
-    MEVehicle* newLeader = 0; // first vehicle in the current queue
-    SUMOTime tleave = MAX2(time + TIME2STEPS(myLength / uspeed) + veh->getStoptime(this) + getLinkPenalty(veh), myBlockTimes[nextQueIndex]);
+    MEVehicle* newLeader = nullptr; // first vehicle in the current queue
+    SUMOTime tleave = MAX2(veh->getStoptime(this, time) + TIME2STEPS(myLength / uspeed) + getLinkPenalty(veh), myBlockTimes[nextQueIndex]);
+    if (veh->isStopped()) {
+        MSNet::getInstance()->getVehicleControl().addWaiting(&myEdge, veh);
+    }
     myEdge.lock();
     if (cars.empty()) {
         cars.push_back(veh);
@@ -539,6 +550,7 @@ MESegment::receive(MEVehicle* veh, SUMOTime time, bool isDepart, bool afterTelep
         }
     }
     myEdge.unlock();
+    myNumCars++;
     if (!isDepart) {
         // regular departs could take place anywhere on the edge so they should not block regular flow
         // the -1 facilitates interleaving of multiple streams
@@ -556,7 +568,7 @@ MESegment::receive(MEVehicle* veh, SUMOTime time, bool isDepart, bool afterTelep
     } else {
         veh->activateReminders(MSMoveReminder::NOTIFICATION_SEGMENT);
     }
-    if (newLeader != 0) {
+    if (newLeader != nullptr) {
         MSGlobals::gMesoNet->addLeaderCar(newLeader, getLink(newLeader));
     }
 }
@@ -564,7 +576,7 @@ MESegment::receive(MEVehicle* veh, SUMOTime time, bool isDepart, bool afterTelep
 
 bool
 MESegment::vaporizeAnyCar(SUMOTime currentTime) {
-    MEVehicle* remove = 0;
+    MEVehicle* remove = nullptr;
     for (Queues::const_iterator k = myCarQues.begin(); k != myCarQues.end(); ++k) {
         if (!k->empty()) {
             // remove last in queue
@@ -651,10 +663,13 @@ void
 MESegment::loadState(std::vector<std::string>& vehIds, MSVehicleControl& vc, const SUMOTime block, const int queIdx) {
     for (std::vector<std::string>::const_iterator it = vehIds.begin(); it != vehIds.end(); ++it) {
         MEVehicle* v = static_cast<MEVehicle*>(vc.getVehicle(*it));
-        assert(v != 0);
-        assert(v->getSegment() == this);
-        myCarQues[queIdx].push_back(v);
-        myOccupancy += v->getVehicleType().getLengthWithGap();
+        // vehicle could be removed due to options
+        if (v != nullptr) {
+            assert(v->getSegment() == this);
+            myCarQues[queIdx].push_back(v);
+            myNumCars++;
+            myOccupancy += v->getVehicleType().getLengthWithGap();
+        }
     }
     if (myCarQues[queIdx].size() != 0) {
         // add the last vehicle of this queue
@@ -697,7 +712,7 @@ MESegment::getFlow() const {
 SUMOTime
 MESegment::getLinkPenalty(const MEVehicle* veh) const {
     const MSLink* link = getLink(veh, myTLSPenalty || myMinorPenalty);
-    if (link != 0) {
+    if (link != nullptr) {
         SUMOTime result = 0;
         if (link->isTLSControlled()) {
             result += link->getMesoTLSPenalty();
@@ -719,7 +734,7 @@ double
 MESegment::getTLSCapacity(const MEVehicle* veh) const {
     if (myTLSPenalty) {
         const MSLink* link = getLink(veh, true);
-        if (link != 0) {
+        if (link != nullptr) {
             assert(link->isTLSControlled());
             assert(link->getGreenFraction() > 0);
             return link->getGreenFraction();
