@@ -1,13 +1,11 @@
 /****************************************************************************/
 // Eclipse SUMO, Simulation of Urban MObility; see https://eclipse.org/sumo
-// Copyright (C) 2001-2017 German Aerospace Center (DLR) and others.
-/****************************************************************************/
-//
-//   This program and the accompanying materials
-//   are made available under the terms of the Eclipse Public License v2.0
-//   which accompanies this distribution, and is available at
-//   http://www.eclipse.org/legal/epl-v20.html
-//
+// Copyright (C) 2001-2018 German Aerospace Center (DLR) and others.
+// This program and the accompanying materials
+// are made available under the terms of the Eclipse Public License v2.0
+// which accompanies this distribution, and is available at
+// http://www.eclipse.org/legal/epl-v20.html
+// SPDX-License-Identifier: EPL-2.0
 /****************************************************************************/
 /// @file    MSAbstractLaneChangeModel.h
 /// @author  Daniel Krajzewicz
@@ -27,12 +25,9 @@
 // ===========================================================================
 // included modules
 // ===========================================================================
-#ifdef _MSC_VER
-#include <windows_config.h>
-#else
 #include <config.h>
-#endif
 
+#include <microsim/MSGlobals.h>
 #include <microsim/MSVehicle.h>
 
 class MSLane;
@@ -143,6 +138,16 @@ public:
      */
     static MSAbstractLaneChangeModel* build(LaneChangeModel lcm, MSVehicle& vehicle);
 
+    /// @brief whether any kind of lateral dynamics is active
+    inline static bool haveLateralDynamics() {
+        return MSGlobals::gLateralResolution > 0 || MSGlobals::gLaneChangeDuration > 0;
+    }
+
+    /** @brief Returns the model's ID;
+     * @return The model's ID
+     */
+    virtual LaneChangeModel getModelID() const = 0;
+
     /// @brief whether lanechange-output is active
     static bool haveLCOutput() {
         return myLCOutput;
@@ -190,8 +195,12 @@ public:
         return mySavedStates.find(dir)->second;
     }
 
-    void saveState(const int dir, const int stateWithoutTraCI, const int state) {
-        mySavedStates[dir] = std::make_pair(stateWithoutTraCI, state);
+    bool hasSavedState(const int dir) const {
+        return mySavedStates.find(dir) != mySavedStates.end();
+    }
+
+    void saveLCState(const int dir, const int stateWithoutTraCI, const int state) {
+        mySavedStates[dir] = std::make_pair(stateWithoutTraCI | myCanceledStates[dir], state);
     }
 
     void setFollowerGaps(CLeaderDist follower, double secGap);
@@ -202,17 +211,22 @@ public:
     void setOrigLeaderGaps(const MSLeaderDistanceInfo& vehicles);
 
     virtual void prepareStep() {
-        saveState(-1, LCA_UNKNOWN, LCA_UNKNOWN);
-        saveState(0, LCA_UNKNOWN, LCA_UNKNOWN);
-        saveState(1, LCA_UNKNOWN, LCA_UNKNOWN);
+        myCanceledStates[-1] = LCA_NONE;
+        myCanceledStates[0] = LCA_NONE;
+        myCanceledStates[1] = LCA_NONE;
+        saveLCState(-1, LCA_UNKNOWN, LCA_UNKNOWN);
+        saveLCState(0, LCA_UNKNOWN, LCA_UNKNOWN);
+        saveLCState(1, LCA_UNKNOWN, LCA_UNKNOWN);
         myLastLateralGapRight = NO_NEIGHBOR;
         myLastLateralGapLeft = NO_NEIGHBOR;
-        myLastLeaderGap = NO_NEIGHBOR;
-        myLastLeaderSecureGap = NO_NEIGHBOR;
-        myLastFollowerGap = NO_NEIGHBOR;
-        myLastFollowerSecureGap = NO_NEIGHBOR;
-        myLastOrigLeaderGap = NO_NEIGHBOR;
-        myLastOrigLeaderSecureGap = NO_NEIGHBOR;
+        if (!myDontResetLCGaps) {
+            myLastLeaderGap = NO_NEIGHBOR;
+            myLastLeaderSecureGap = NO_NEIGHBOR;
+            myLastFollowerGap = NO_NEIGHBOR;
+            myLastFollowerSecureGap = NO_NEIGHBOR;
+            myLastOrigLeaderGap = NO_NEIGHBOR;
+            myLastOrigLeaderSecureGap = NO_NEIGHBOR;
+        }
         myCommittedSpeed = 0;
     }
 
@@ -313,6 +327,16 @@ public:
     virtual void changed() = 0;
 
 
+    /// @brief return factor for modifying the safety constraints of the car-following model
+    virtual double getSafetyFactor() const {
+        return 1.0;
+    }
+
+    /// @brief return factor for modifying the safety constraints for opposite-diretction overtaking of the car-following model
+    virtual double getOppositeSafetyFactor() const {
+        return 1.0;
+    }
+
     /// @brief whether the current vehicles shall be debugged
     virtual bool debugVehicle() const {
         return false;
@@ -376,14 +400,32 @@ public:
         return myLaneChangeCompletion >= 0.5;
     }
 
-    /// @brief return whether the vehicle passed the midpoint of a continuous lane change maneuver
+    /// @brief Compute the remaining time until LC completion
     SUMOTime remainingTime() const;
 
+    /// @brief Calculates the maximal time needed to complete a lane change maneuver
+    ///        if lcMaxSpeedLatFactor and lcMaxSpeedStanding are set and the vehicle breaks not harder than decel.
+    ///        LC when the vehicle starts breaking now. If lcMaxSpeedStanding==0 the completion may be impossible,
+    /// @param[in] speed Current longitudinal speed of the changing vehicle.
+    /// @param[in] remainingManeuverDist dist which is still to be covered until LC is completed
+    /// @param[in] decel Maximal assumed deceleration rate applied during the LC.
+    /// @return maximal LC duration (or -1) if it is possible that it can't be completed.
+    /// @note 1) For the calculation it is assumed that the vehicle starts breaking with decel (>=0) immediately.
+    ///       If lcMaxSpeedStanding==0 the completion may be impossible, and -1 is returned.
+    ///       2) In case that no maxSpeedLat is used to control lane changing, this is only called prior to a lane change,
+    ///          and the duration is MSGlobals::gLaneChangeDuration.
+    virtual double estimateLCDuration(const double speed, const double remainingManeuverDist, const double decel) const;
+
     /// @brief return true if the vehicle currently performs a lane change maneuver
     inline bool isChangingLanes() const {
         return myLaneChangeCompletion < (1 - NUMERICAL_EPS);
     }
 
+    /// @brief Get the current lane change completion ratio
+    inline double getLaneChangeCompletion() const {
+        return myLaneChangeCompletion;
+    }
+
     /// @brief return the direction of the current lane change maneuver
     inline int getLaneChangeDirection() const {
         return myLaneChangeDirection;
@@ -408,6 +450,10 @@ public:
     /// @brief start the lane change maneuver and return whether it continues
     bool startLaneChangeManeuver(MSLane* source, MSLane* target, int direction);
 
+    /// @brief Control for resetting the memorized values for LC relevant gaps until the LC output is triggered in the case of continuous LC.
+    void memorizeGapsAtLCInit();
+    void clearGapsAtLCInit();
+
     /* @brief continue the lane change maneuver and return whether the midpoint
      * was passed in this step
      */
@@ -484,11 +530,11 @@ public:
 
     /// @brief decides the next lateral speed depending on the remaining lane change distance to be covered
     ///        and updates maneuverDist according to lateral safety constraints.
-    virtual double computeSpeedLat(double latDist, double& maneuverDist) {
-        UNUSED_PARAMETER(latDist);
-        UNUSED_PARAMETER(maneuverDist);
-        throw ProcessError("Method not implemented by model " + toString(myModel));
-    }
+    virtual double computeSpeedLat(double latDist, double& maneuverDist);
+
+    /// @brief Returns a deceleration value which is used for the estimation of the duration of a lane change.
+    /// @note  Effective only for continuous lane-changing when using attributes myMaxSpeedLatFactor and myMaxSpeedLatStanding. See #3771
+    virtual double getAssumedDecelForLaneChangeDuration() const;
 
     /// @brief try to retrieve the given parameter from this laneChangeModel. Throw exception for unsupported key
     virtual std::string getParameter(const std::string& key) const {
@@ -514,7 +560,7 @@ protected:
     virtual bool predInteraction(const std::pair<MSVehicle*, double>& leader);
 
     /// @brief whether the influencer cancels the given request
-    bool cancelRequest(int state);
+    bool cancelRequest(int state, int laneOffset);
 
 
 protected:
@@ -529,6 +575,7 @@ protected:
     int myPreviousState2;
 
     std::map<int, std::pair<int, int> > mySavedStates;
+    std::map<int, int> myCanceledStates;
 
     /// @brief the current lateral speed
     double mySpeedLat;
@@ -601,6 +648,15 @@ protected:
     double myLastOrigLeaderGap;
     double myLastOrigLeaderSecureGap;
 
+    /// @brief Flag to prevent resetting the memorized values for LC relevant gaps until the LC output is triggered
+    ///        in the case of continuous LC.
+    bool myDontResetLCGaps;
+
+    // @brief the maximum lateral speed when standing
+    double myMaxSpeedLatStanding;
+    // @brief the factor of maximum lateral speed to longitudinal speed
+    double myMaxSpeedLatFactor;
+
     /* @brief to be called by derived classes in their changed() method.
      * If dir=0 is given, the current value remains unchanged */
     void initLastLaneChangeOffset(int dir);
